#pragma once

#include "distributions.hpp"
#include "pde.hpp"
#include <vector>
#include <queue>

namespace wos {

enum class Algorithm {
	ConstantCoeffs,
	DeltaTracking
};

struct WeightWindow {
	// constructor
	WeightWindow(float lowerBound_, float upperBound_): lowerBound(lowerBound_), upperBound(upperBound_) {}

	// members
	float lowerBound;
	float upperBound;
};

template <typename T, int DIM>
struct WalkState;

// NOTE: use Eigen::Array in place of Eigen::VectorXf as it does not support component wise operations
template <typename T, int DIM>
class SampleStatistics;

template <typename T, int DIM>
class WalkOnSpheres {
public:
	// constructor
	WalkOnSpheres(const std::function<float(const Vector<DIM>&)>& distToBoundary_, int maxWalkLength_):
								distToBoundary(distToBoundary_), maxWalkLength(maxWalkLength_) {
	  // nothing to do
	}

	// solves an elliptic PDE at a single input point
	void solve(const Vector<DIM>& pt, const PDE<T, DIM>& pde,
						 Algorithm algorithm, WeightWindow window, int nWalks, float epsilonShell,
						 T initVal, SampleStatistics<T, DIM>& statistics) const {
		// choose a greens function based on the algorithm and medium
		std::unique_ptr<GreensFnBall<DIM>> greensFn = nullptr;
		if (algorithm == Algorithm::ConstantCoeffs) {
			if (pde.mediumType == MediumType::Empty) {
				greensFn = std::make_unique<HarmonicGreensFnBall<DIM>>();

			} else {
				greensFn = std::make_unique<YukawaGreensFnBall<DIM>>(pde.bound);
			}

		} else {
			greensFn = std::make_unique<YukawaGreensFnBall<DIM>>(pde.bound);
		}


		float initDist = distToBoundary(pt);
		// run no walks if we are outside the domain
		if (initDist < 0) {
			nWalks = 0;
		}
		// run a single random walk if the distance to the boundary is smaller than the epsilon shell
		else if (initDist <= epsilonShell) {
			nWalks = 1;
		}

		// perform random walks
		std::queue<WalkState<T, DIM>> queue;
		for (int w = 0; w < nWalks; w++) {
			// initialize the walk state
			WalkState<T, DIM> state(pt, 1.0f, 0, false, initVal);
			T totalContribution = initVal;
			bool anySuccess = false;
			int splitsPerformed = -1;
			queue.emplace(state);

			while (!queue.empty()) {
				// perform a walk with the dequeued state
				state = queue.front();
				queue.pop();
				splitsPerformed++;
				bool success = true;

				if (algorithm == Algorithm::ConstantCoeffs) {
					success = walkConstantCoeffs(pde, window, epsilonShell, initVal, greensFn, state, queue);
				} else if (algorithm == Algorithm::DeltaTracking) {
					success = walkDeltaTracking(pde, window, epsilonShell, initVal, greensFn, state, queue);
				}

				if (success) {
					// get the boundary contribution
					T boundaryContribution = pde.dirichlet(state.currentPt);

					// update the total walk contribution
					totalContribution += state.potential*boundaryContribution + state.totalSourceContribution;

					// record the walk length
					statistics.addWalkLength(state.walkLength);
					anySuccess = true;
				}
			}

			if (anySuccess) {
				// update the solution estimate
				statistics.addSolutionContribution(totalContribution);

				// record the number of splits
				statistics.addSplits(splitsPerformed);
			}
		}
	}

	// solves an elliptic PDE at the input points
	void solve(const std::vector<Vector<DIM>>& pts, const PDE<T, DIM>& pde,
						 Algorithm algorithm, WeightWindow window, int nWalks, float epsilonShell,
						 T initVal, std::vector<SampleStatistics<T, DIM>>& statistics) const {
		// initialize a statistics object for each point if required
		int nPoints = (int)pts.size();
		if (statistics.size() == 0) {
			statistics.resize(nPoints, SampleStatistics<T, DIM>(initVal));
		}

		#pragma omp parallel for
		for (int i = 0; i < nPoints; ++i) {
			solve(pts[i], pde, algorithm, window, nWalks, epsilonShell, initVal, statistics[i]);
		}
	}

private:
	// terminates, splits or continues a walk based on its potential value
	bool applyWeightWindow(WeightWindow window, T initVal, WalkState<T, DIM>& state,
												 std::queue<WalkState<T, DIM>>& queue) const {
    if (window.upperBound > window.lowerBound) {
			if (state.potential > window.upperBound) {
				// split the walk
				float potentialLeft = state.potential - window.upperBound;
				state.potential = window.upperBound;
				WalkState<T, DIM> splitState(state.currentPt, state.potential, 0, true, initVal);

				while (potentialLeft > window.upperBound) {
					potentialLeft -= window.upperBound;
					queue.emplace(splitState);
				}

				if (sampleUniform() < potentialLeft/window.upperBound) {
					queue.emplace(splitState);
				}

			} else if (state.potential < window.lowerBound) {
				// kill the walk using russian roulette
				float survivalProb = state.potential/window.lowerBound;
				if (survivalProb < sampleUniform()) {
					state.potential = 0.0f;
					return true;
				}

				state.potential = window.lowerBound;
			}
		}

		return false;
	}

	// performs a single random walk starting at the input point for a constant coeff problem.
	// returns false if the walk does not terminate
	bool walkConstantCoeffs(const PDE<T, DIM>& pde, WeightWindow window, float epsilonShell,
													T initVal, std::unique_ptr<GreensFnBall<DIM>>& greensFn,
													WalkState<T, DIM>& state, std::queue<WalkState<T, DIM>>& queue) const {
		// recursively perform a random walk till it reaches the boundary
		float dist = distToBoundary(state.currentPt);
		float diffusion = pde.diffusion(state.currentPt);

		while (dist > epsilonShell) {
			// update the sphere center and radius
			greensFn->updateBall(state.currentPt, dist);

			// compute the source contribution inside the sphere
			float pdf;
			greensFn->sampleVolume(pdf);
			float greensFnNorm = greensFn->norm();
			T sourceContribution = greensFnNorm*pde.source(greensFn->yVol)/diffusion;
			state.totalSourceContribution += state.potential*sourceContribution;

			// sample a point uniformly on the sphere, and update the current
			// position of the walk
			state.currentPt = greensFn->sampleSurface(pdf);

			// update the potential
			state.potential *= greensFn->poissonKernel()/pdf;

			// apply the weight window
			bool terminate = applyWeightWindow(window, initVal, state, queue);
			if (terminate) return true;

			// compute the distance of the currentPt to the boundary
			dist = distToBoundary(state.currentPt);

			// update the walk length and break if the max walk length is exceeded
			state.walkLength++;
			if (state.walkLength > maxWalkLength) {
				std::cerr << "Maximum walk length exceeded!" << std::endl;
				break;
			}
		}

		return dist <= epsilonShell;
	}

	// performs a single random walk starting at the input point for a variable coeff problem;
	// returns false if the walk does not terminate
	bool walkDeltaTracking(const PDE<T, DIM>& pde, WeightWindow window, float epsilonShell,
												 T initVal, std::unique_ptr<GreensFnBall<DIM>>& greensFn,
												 WalkState<T, DIM>& state, std::queue<WalkState<T, DIM>>& queue) const {
		// recursively perform a random walk till it reaches the boundary
		float dist = distToBoundary(state.currentPt);
		float diffusion = pde.diffusion(state.currentPt);

		while (dist > epsilonShell) {
			// update the sphere center and radius
			greensFn->updateBall(state.currentPt, dist);

			// compute the source contribution inside the sphere
			float pdf;
			greensFn->sampleVolume(pdf);
			float greensFnNorm = greensFn->norm();
			float diffusionWeight = 1.0f/std::sqrt(diffusion*pde.diffusion(greensFn->yVol));
			T sourceContribution = greensFnNorm*diffusionWeight*pde.source(greensFn->yVol);
			state.totalSourceContribution += state.potential*sourceContribution;

			// decide whether to sample the volume or boundary integral
			bool sampleVolumeIntegral = sampleUniform() < pde.bound*greensFnNorm;
			if (sampleVolumeIntegral) {
				// sample a point inside the sphere using the centered greens function,
				// and update the current position of the walk
				float pdf;
				state.currentPt = greensFn->sampleVolume(pdf);

				// update the potential with the collision probability
				state.potential *= (1.0f - pde.transformedAbsorption(state.currentPt)/pde.bound);

			} else {
				// sample a point uniformly on the sphere, and update the current
				// position of the walk
				float pdf;
				state.currentPt = greensFn->sampleSurface(pdf);
			}

			// update the potential with the diffusion weight
			float diffusionCurrentPt = pde.diffusion(state.currentPt);
			state.potential *= std::sqrt(diffusionCurrentPt/diffusion);
			diffusion = diffusionCurrentPt;

			// apply the weight window
			bool terminate = applyWeightWindow(window, initVal, state, queue);
			if (terminate) return true;

			// compute the distance of the currentPt to the boundary
			dist = distToBoundary(state.currentPt);

			// update the walk length and break if the max walk length is exceeded
			state.walkLength++;
			if (state.walkLength > maxWalkLength) {
				std::cerr << "Maximum walk length exceeded!" << std::endl;
				break;
			}
		}

		return dist <= epsilonShell;
	}

	// members
	std::function<float(const Vector<DIM>&)> distToBoundary;
	int maxWalkLength;
};

template <typename T, int DIM>
struct WalkState {
	// constructor
	WalkState(const Vector<DIM>& currentPt_, float potential_,
						int walkLength_, bool isSplitState_, T initVal):
						currentPt(currentPt_),
						totalSourceContribution(initVal),
						potential(potential_),
						walkLength(walkLength_),
						isSplitState(isSplitState_) {}

	// members
	Vector<DIM> currentPt;
	T totalSourceContribution;
	float potential;
	int walkLength;
	bool isSplitState;
};

template <typename T, int DIM>
class SampleStatistics {
public:
	// constructor
	SampleStatistics(T initVal) {
		reset(initVal);
	}

	// resets statistics
	void reset(T initVal) {
		solutionMean = initVal;
		solutionM2 = initVal;
		for (int i = 0; i < DIM; i++) {
			gradientMean[i] = initVal;
			gradientM2[i] = initVal;
		}
		firstSourceContributionSum = initVal;
		nSolutionContributions = 0;
		nGradientContributions = 0;
		totalWalkLength = 0;
		totalSplits = 0;
	}

	// adds solution contribution to running sum
	void addSolutionContribution(const T& contribution) {
		nSolutionContributions += 1;
		update(contribution, solutionMean, solutionM2, nSolutionContributions);
	}

	// adds gradient contribution to running sum
	void addGradientContribution(const T *boundaryContribution, const T *sourceContribution) {
		nGradientContributions += 1;
		for (int i = 0; i < DIM; i++) {
			update(boundaryContribution[i] + sourceContribution[i],
						 gradientMean[i], gradientM2[i], nGradientContributions);
		}
	}

	// add first source contribution
	void addFirstSourceContribution(const T& contribution) {
		firstSourceContributionSum += contribution;
	}

	// adds walk length to running sum
	void addWalkLength(int length) {
		totalWalkLength += length;
	}

	// adds walk length to running sum
	void addSplits(int nSplits) {
		totalSplits += nSplits;
	}

	// returns the solution estimate
	T getSolution() const {
		return solutionMean;
	}

	// returns the estimate of the solution variance
	T getSolutionVariance() const {
		int N = std::max(1, nSolutionContributions - 1);
		return solutionM2/N;
	}

	// returns the gradient estimate
	const T* getGradient() const {
		return gradientMean;
	}

	// returns the estimate of the gradient variance
	std::vector<T> getGradientVariance() const {
		int N = std::max(1, nGradientContributions - 1);
		std::vector<T> variance(DIM);

		for (int i = 0; i < DIM; i++) {
			variance[i] = gradientM2[i]/N;
		}

		return variance;
	}

	// returns first source contribution
	T getFirstSourceContribution() const {
		int N = std::max(1, nSolutionContributions);
		return firstSourceContributionSum/N;
	}

	// returns the number of solution contributions
	int getSolutionContributionsCount() const {
		return nSolutionContributions;
	}

	// returns the number of gradient contributions
	int getGradientContributionsCount() const {
		return nGradientContributions;
	}

	// returns the mean walk length
	float getMeanWalkLength() const {
		int N = std::max(1, nSolutionContributions);
		return (float)totalWalkLength/N;
	}

	// returns the total number of splits
	int getTotalSplits() const {
		return totalSplits;
	}

private:
	// updates statistics
	void update(const T& contribution, T& mean, T& M2, int N) {
		T delta = contribution - mean;
		mean += delta/N;
		T delta2 = contribution - mean;
		M2 += delta*delta2;
	}

	// members
	T solutionMean, solutionM2;
	T gradientMean[DIM], gradientM2[DIM];
	T firstSourceContributionSum;
	int nSolutionContributions, nGradientContributions;
	int totalWalkLength, totalSplits;
};

} // mc
